#include "logger.h"

int pyfcgi_logger_add(const char *filename, logmask_t loglvl, logmask_t logtyp,
                      const char *format)
{
	pyfcgi_conf_logger_t *conf;
	pyfcgi_logger_t logger;
	size_t new_idx;
	void *tmp;
	int err;
	char *err_fmt;

	conf = &PyFCGI_conf.logs;

	if(conf->logger_sz == 255)
	{
		pyfcgi_log(LOG_ERR, "Maximum of 255 logger reached, unable to add this one...");
		return PYFCGI_ERR;
	}

	if(pyfcgi_logger_add_format(format, &(logger.fmt_id)))
	{
		return PYFCGI_FATAL;
	}
	if(!logger.filename)
	{
		err_fmt = "Unable to duplicate logger filename : %s";
		goto err;
	}

	logger.loglvl = loglvl;
	logger.logtyp = logtyp;
	logger.filename = strdup(filename);

	new_idx = conf->logger_sz;
	conf->logger_sz++;
	if( !(tmp = realloc(conf->loggers,
	                    sizeof(pyfcgi_logger_t) * conf->logger_sz)) )
	{
		err = errno;
		err_fmt = "Unable to realloc loggers array : %s";
		goto err_free_filename;
	}
	conf->loggers[new_idx] = logger;
	return 0;

err_free_filename:
	free(logger.filename);
err:
	pyfcgi_log(LOG_ALERT,
	           err_fmt,
		   strerror(err));
	return PYFCGI_FATAL;	
}

int pyfcgi_logger_add_format(const char* format, size_t* idx)
{
	char *fmt;
	size_t i;
	void *tmp;
	pyfcgi_conf_logger_t *conf;

	conf = &PyFCGI_conf.logs;

	for(i=0; i<conf->format_sz; i++)
	{
		if(!strcmp(format, conf->format[i])) { break; }
	}
	if(idx) { *idx = i; }
	if( i<conf->format_sz )
	{
		return 0;
	}
	if( !(fmt = strdup(format)) )
	{
		pyfcgi_log(LOG_ALERT, "Fails to strdup new format : %s",
			   strerror(errno));
		return PYFCGI_FATAL;
	}
	conf->format_sz++;
	if( !(tmp = realloc(conf->format,
	                    sizeof(char*) * conf->format_sz)) )
	{
		conf->format_sz--;
		pyfcgi_log(LOG_ALERT, "Fails to realloc logger format array : %s",
			   strerror(errno));
		return PYFCGI_FATAL;
	}
	conf->format[i] = fmt;
	return 0;
}

int pyfcgi_logger_open(pyfcgi_logger_t *logger)
{
	if( (logger->fd = open(logger->filename,
	                       O_WRONLY | O_APPEND | O_CREAT, 00640) < 0) )
	{
		pyfcgi_log(LOG_ERR,
		           "Unable to open log file '%s' : %s",
			   strerror(errno));
		return PYFCGI_ERR;
	}
	return 0;
}

int pyfcgi_log(loglvl_t lvl, const char *fmt, ...)
{
	int ret;
	va_list ap;
	va_start(ap, fmt);
	ret = vpyfcgi_log(lvl, fmt, ap);
	va_end(ap);
	return ret;
}

int vpyfcgi_log(loglvl_t lvl, const char *fmt, va_list ap)
{
	int len, ret;
	pyfcgi_conf_logger_t *conf;
	unsigned char i;
	char buf[512];

	conf = &PyFCGI_conf.logs;

	_vsyslog(lvl, fmt, ap);

	len = 0;
	if(conf->format_sz > 1)
	{
		len = vdprintf(conf->pipes[1], fmt, ap);
		if(len < 0)
		{
			_syslog(LOG_ALERT, 
			       "Unable to write to multiplexer pipe when trying to log : %s",
			       strerror(errno));
			return PYFCGI_FATAL;
		}
	}
	else if (conf->format_sz)
	{
		len = vdprintf(conf->loggers[0].fd, fmt, ap);
		if(len < 0)
		{
			_syslog(LOG_ALERT, 
			       "Unable to write to single FD to '%s' when trying to log : %s",
			       conf->loggers[0].filename, strerror(errno));
			return PYFCGI_FATAL;
		}
		return 0;
	}
	for(i=0; i<conf->logger_sz-1; i++)
	{
		ret = tee(conf->pipes[0], conf->loggers[i].fd, len, 0);
		if(ret < 0)
		{
			_syslog(LOG_ALERT, 
			       "Unable to splice to last logfile '%s' : %s",
			       conf->loggers[i].filename, strerror(errno));
			return PYFCGI_FATAL;
		}
	}
	ret = splice(conf->pipes[0], NULL, conf->loggers[i].fd, NULL, len, 0);
	if(ret < 0)
	{
		_syslog(LOG_ALERT, 
		       "Unable to splice to last logfile '%s' : %s",
		       conf->loggers[i].filename, strerror(errno));
		return PYFCGI_ERR;
	}
	if(ret < len)
	{
		_syslog(LOG_WARNING, 
		       "Unable to splice all data to logfile. Flushing pipe.");

		do
		{
			len -= (ret<len)?ret:len;
			ret = read(conf->pipes[0], buf, (len>sizeof(buf))?sizeof(buf):len);
			if(ret < 0)
			{
				_syslog(LOG_CRIT,
				        "Logger pipe seems broken, unable to flush : %s",
					strerror(errno));
				return PYFCGI_ERR;
			}
		}while(len);
	}
	return 0;
}
